from django.db import models
from django.utils.translation import gettext_lazy as _


class AddressMixin(models.Model):
    """
    Base model for addresses, containing common fields used in all address types.
    """

    title = models.CharField(
        _("Title"),
        max_length=100,
        help_text=_("A descriptive title for the address, such as 'Home' or 'Office'."),
        db_comment="Descriptive title for the address.",
    )
    country = models.CharField(
        _("Country"),
        max_length=100,
        help_text=_("The country where the address is located."),
        db_comment="Country of the address.",
    )
    province = models.CharField(
        _("Province"),
        max_length=100,
        help_text=_("The province, state, or regional division of the address."),
        db_comment="Province or state of the address.",
    )
    city = models.CharField(
        _("City"),
        max_length=100,
        help_text=_("The city or town of the address."),
        db_comment="City or town of the address.",
    )
    postal_address = models.TextField(
        _("Postal Address"),
        help_text=_(
            "The full postal address, including street name, number, and any additional details."
        ),
        db_comment="Full detailed postal address.",
    )
    postal_code = models.CharField(
        _("Postal Code"),
        max_length=20,
        help_text=_("The postal or ZIP code for the address."),
        db_comment="Postal or ZIP code of the address.",
    )
    plaque = models.CharField(
        _("Plaque"),
        max_length=20,
        help_text=_("The specific plaque number of the building or house."),
        db_comment="Plaque number of the building or house.",
    )
    building_unit = models.CharField(
        _("Building Unit"),
        max_length=20,
        blank=True,
        null=True,
        help_text=_("The unit or apartment number within a building, if applicable."),
        db_comment="Unit or apartment number in a building, optional.",
    )

    class Meta:
        abstract = True



import logging
from django.conf import settings

from typing import Any
from django.utils import timezone
from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin

from sage_tools.handlers.session import SessionHandler

logger = logging.getLogger(__name__)

try:
    import pytz
except ImportError:
    raise ImportError(  # noqa: B904
        "Install `pytz` package. Run `pip install pytz`."
    )

class TimezoneMiddleware(MiddlewareMixin):
    """
    Middleware to handle setting the user's timezone based on their session data.
    """

    def process_request(self, request: HttpRequest) -> None:
        """
        Process the request to set the timezone from the session.
        """
        session_handler = SessionHandler(request)
        tzname = session_handler.get('user_timezone')
        if tzname:
            try:
                timezone.activate(pytz.timezone(tzname))
                # Set Django's timezone to the user's timezone
                settings.TIME_ZONE = tzname
            except pytz.UnknownTimeZoneError:
                logger.error(f"Unknown timezone: {tzname}")
                timezone.deactivate()
        else:
            timezone.deactivate()

    def process_response(self, request: HttpRequest, response) -> Any:
        """
        Ensure the timezone is deactivated after the response is processed.
        """
        timezone.deactivate()
        return response



from .base import Encryptor


class DummyEncryptor(Encryptor):
    """
    Dummy encryption class for testing.

    This class implements the Encryptor interface but performs no actual encryption or decryption.

    Methods
    -------
    encrypt(data: str) -> str
        Returns the data unchanged.
    decrypt(data: str) -> str
        Returns the data unchanged.

    Examples
    --------
    >>> encryptor = DummyEncryptor()
    >>> encrypted_data = encryptor.encrypt("Hello, World!")
    >>> encrypted_data
    'Hello, World!'
    >>> decryptor = encryptor.decrypt(encrypted_data)
    >>> decryptor
    'Hello, World!'
    """

    def encrypt(self, data: str) -> str:
        return data

    def decrypt(self, data: str) -> str:
        return data



"""
This module is derived from the django-braces package.
"""

import datetime
import inspect
import urllib.parse

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import logout_then_login, redirect_to_login
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import (
    Http404,
    HttpResponse,
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
    StreamingHttpResponse,
)
from django.shortcuts import resolve_url
from django.utils.encoding import force_str
from django.utils.timezone import now


class AccessMixin:
    """
    Base access mixin. All other access mixins should extend this one.
    """

    login_url = None
    raise_exception = False
    redirect_field_name = REDIRECT_FIELD_NAME  # Set by django.contrib.auth
    redirect_unauthenticated_users = False

    def __init__(self, *args, **kwargs):
        self._class_name = self.__class__.__name__
        super().__init__(*args, **kwargs)

    def get_login_url(self):
        """
        Override this method to customize the login_url.
        """
        login_url = self.login_url or settings.LOGIN_URL
        if not login_url:
            raise ImproperlyConfigured(
                f"Define {self._class_name}.login_url or settings.LOGIN_URL or "
                f"override {self._class_name}.get_login_url()."
            )

        return force_str(login_url)

    def get_redirect_field_name(self):
        """
        Override this method to customize the redirect_field_name.
        """
        if self.redirect_field_name is None:
            raise ImproperlyConfigured(
                f"{self._class_name} is missing the redirect_field_name. "
                f"Define {self._class_name}.redirect_field_name or "
                f"override {self._class_name}.get_redirect_field_name()."
            )
        return self.redirect_field_name

    def handle_no_permission(self, request):
        """What should happen if the user doesn't have permission?"""
        if self.raise_exception:
            if (
                self.redirect_unauthenticated_users
                and not request.user.is_authenticated
            ):
                return self.no_permissions_fail(request)
            else:
                if inspect.isclass(self.raise_exception) and issubclass(
                    self.raise_exception, Exception
                ):
                    raise self.raise_exception
                if callable(self.raise_exception):
                    ret = self.raise_exception(request)
                    if isinstance(ret, (HttpResponse, StreamingHttpResponse)):
                        return ret
                raise PermissionDenied

        return self.no_permissions_fail(request)

    def no_permissions_fail(self, request=None):
        """
        Called when the user has no permissions and no exception was raised.
        This should only return a valid HTTP response.

        By default we redirect to login.
        """
        return redirect_to_login(
            request.get_full_path(),
            self.get_login_url(),
            self.get_redirect_field_name(),
        )


class LoginRequiredMixin(AccessMixin):
    """
    Requires the user to be authenticated.

    NOTE:
        This should be the left-most mixin of a view, except when
        combined with CsrfExemptMixin - which in that case should
        be the left-most mixin.
    """

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate method after checking authentication"""
        if not request.user.is_authenticated:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class AnonymousRequiredMixin(AccessMixin):
    """
    Requires the user to be unauthenticated.

    NOTE:
        This should be the left-most mixin of a view.

    ## Example Usage

        class SomeView(AnonymousRequiredMixin, ListView):
            ...
            # required
            authenticated_redirect_url = "/accounts/profile/"
            ...
    """

    authenticated_redirect_url = settings.LOGIN_REDIRECT_URL

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler after guaranteeing anonymity"""
        if request.user.is_authenticated:
            return HttpResponseRedirect(self.get_authenticated_redirect_url())
        return super().dispatch(request, *args, **kwargs)

    def get_authenticated_redirect_url(self):
        """Return the reversed authenticated redirect url."""
        if not self.authenticated_redirect_url:
            raise ImproperlyConfigured(
                f"{self._class_name} is missing an authenticated_redirect_url "
                f"url to redirect to. Define {self._class_name}.authenticated_redirect_url "
                f"or override {self._class_name}.get_authenticated_redirect_url()."
            )
        return resolve_url(self.authenticated_redirect_url)


class PermissionRequiredMixin(AccessMixin):
    """
    The request users must have certain permission(s)

    ## Attributes

    `permission_required` - the permission to check for.
    `login_url` - the login url of site
    `redirect_field_name` - defaults to "next"
    `raise_exception` - defaults to False - raise 403 if set to True

    ## Example Usage

        class SomeView(PermissionRequiredMixin, ListView):
            ...
            # required
            permission_required = "app.permission"

            # optional
            login_url = "/signup/"
            redirect_field_name = "hollaback"
            raise_exception = True
            ...
    """

    permission_required = None  # No permissions are required by default
    object_level_permissions = False

    def get_permission_required(self, request=None):
        """
        Get the required permissions and return them.

        Override this to allow for custom permission_required values.
        """
        # Make sure that the permission_required attribute is set on the
        # view, or raise a configuration error.
        if self.permission_required is None:
            raise ImproperlyConfigured(
                f'{self._class_name} requires the "permission_required" '
                "attribute to be set."
            )

        return self.permission_required

    def check_permissions(self, request):
        """
        Returns whether or not the user has permissions
        """
        perms = self.get_permission_required(request)
        has_permission = False

        if self.object_level_permissions:
            if hasattr(self, "object") and self.object is not None:
                has_permission = request.user.has_perm(
                    self.get_permission_required(request), self.object
                )
            elif hasattr(self, "get_object") and callable(self.get_object):
                has_permission = request.user.has_perm(
                    self.get_permission_required(request), self.get_object()
                )
        else:
            has_permission = request.user.has_perm(
                self.get_permission_required(request)
            )
        return has_permission

    def dispatch(self, request, *args, **kwargs):
        """
        Check to see if the user in the request has the required
        permission.
        """
        has_permission = self.check_permissions(request)

        if not has_permission:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class MultiplePermissionsRequiredMixin(PermissionRequiredMixin):
    """
    View mixin which allows you to specify two types of permission
    requirements. The `permissions` attribute must be a dict which
    specifies two keys, `all` and `any`. You can use either one on
    its own or combine them. The value of each key is required to be a
    list or tuple of permissions. The standard Django permissions
    style is not strictly enforced. If you have created your own
    permissions in a different format, they should still work.

    By specifying the `all` key, the user must have all of
    the permissions in the list.

    By specifying the `any` key , the user must have at least one of the
    permissions in the list.

    Class Settings
        `permissions` - This is required to be a dict with one or both
            keys of `all` and `any` containing a list or tuple of
            permissions.
        `login_url` - the login url of site
        `redirect_field_name` - defaults to "next"
        `raise_exception` - defaults to False - raise 403 if set to True

    Example Usage
        class SomeView(MultiplePermissionsRequiredMixin, ListView):
            ...
            #required
            permissions = {
                "all": ("blog.add_post", "blog.change_post"),
                "any": ("blog.delete_post", "user.change_user")
            }

            #optional
            login_url = "/signup/"
            redirect_field_name = "hollaback"
            raise_exception = True
    """

    permissions = None  # Default required perms to none

    def get_permission_required(self, request=None):
        """Get which permission is required"""
        self._check_permissions_attr()
        return self.permissions

    def check_permissions(self, request):
        """Get the permissions, both all and any."""
        permissions = self.get_permission_required(request)
        perms_all = permissions.get("all")
        perms_any = permissions.get("any")

        self._check_permissions_keys_set(perms_all, perms_any)
        self._check_perms_keys("all", perms_all)
        self._check_perms_keys("any", perms_any)

        # Check that user has all permissions in the list/tuple
        if perms_all:
            # Why not `return request.user.has_perms(perms_all)`?
            # There may be optional permissions below.
            if not request.user.has_perms(perms_all):
                return False

        # If perms_any, check that user has at least one in the list/tuple
        if perms_any:
            any_perms = [request.user.has_perm(perm) for perm in perms_any]
            if not any_perms or not any(any_perms):
                return False
        return True

    def _check_permissions_attr(self):
        """
        Check permissions attribute is set and that it is a dict.
        """
        if self.permissions is None or not isinstance(self.permissions, dict):
            raise ImproperlyConfigured(
                f"{self._class_name} requires the `permissions` attribute"
                "to be set as a dict."
            )

    def _check_permissions_keys_set(self, perms_all=None, perms_any=None):
        """
        Check to make sure the keys `any` or `all` are not both blank.
        If both are blank either an empty dict came in or the wrong keys
        came in. Both are invalid and should raise an exception.
        """
        if perms_all is None and perms_any is None:
            raise ImproperlyConfigured(
                f"{self._class_name} requires the `permissions` attribute to "
                f"be set to a dict and the `any` or `all` key to be set."
            )

    def _check_perms_keys(self, key=None, perms=None):
        """
        If the permissions list/tuple passed in is set, check to make
        sure that it is of the type list or tuple.
        """
        if perms and not isinstance(perms, (list, tuple)):
            raise ImproperlyConfigured(
                f"{self._class_name} requires the permissions dict {key} value "
                "to be a list or tuple."
            )


class GroupRequiredMixin(AccessMixin):
    group_required = None

    def get_group_required(self):
        """Get which group's membership is required"""
        if any(
            [
                self.group_required is None,
                not isinstance(self.group_required, (list, tuple, str)),
            ]
        ):

            raise ImproperlyConfigured(
                f"{self._class_name} requires the `group_required` attribute "
                "to be set and be a string, list, or tuple."
            )
        if not isinstance(self.group_required, (list, tuple)):
            self.group_required = (self.group_required,)
        return self.group_required

    def check_membership(self, groups):
        """Check for user's membership in required groups. Superusers are
        automatically members"""
        if self.request.user.is_superuser:
            return True

        user_groups = self.request.user.groups.values_list("name", flat=True)
        return set(groups).intersection(set(user_groups))

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler if the user is a group member"""
        self.request = request
        in_group = False
        if request.user.is_authenticated:
            in_group = self.check_membership(self.get_group_required())

        if not in_group:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class UserPassesTestMixin(AccessMixin):
    """
    User must pass a test before being allowed access to the view.

    Class Settings
        `test_func` - This is required to be a method that takes user
            instance and return True or False after checking conditions.
        `login_url` - the login url of site
        `redirect_field_name` - defaults to "next"
        `raise_exception` - defaults to False - raise 403 if set to True
    """

    def test_func(self, user):
        """The function to test the user with"""
        raise NotImplementedError(
            f"{self._class_name} is missing implementation of the "
            "`test_func` method. A function to test the user is required."
        )

    def get_test_func(self):
        """Get the test function"""
        return getattr(self, "test_func")

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler if the users passes the test"""
        user_test_result = self.get_test_func()(request.user)

        if not user_test_result:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class SuperuserRequiredMixin(AccessMixin):
    """
    Require users to be superusers to access the view.
    """

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler if the user is a superuser"""
        if not request.user.is_superuser:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class StaffuserRequiredMixin(AccessMixin):
    """
    Require users to be marked as staff to access the view.
    """

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler if the user is a staff member"""
        if not request.user.is_staff:
            return self.handle_no_permission(request)

        return super().dispatch(request, *args, **kwargs)


class SSLRequiredMixin:
    """
    Require requests to be made over a secure connection.
    """

    raise_exception = False

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate handler if the connection is secure"""
        if getattr(settings, "DEBUG", False):
            # Don't enforce the check during development
            return super().dispatch(request, *args, **kwargs)

        if not request.is_secure():
            if self.raise_exception:
                raise Http404

            return HttpResponsePermanentRedirect(self._build_https_url(request))

        return super(SSLRequiredMixin, self).dispatch(request, *args, **kwargs)

    def _build_https_url(self, request):
        """Get the full url, replace http with https"""
        url = request.build_absolute_uri(request.get_full_path())
        return urllib.parse.urlunsplit(("https",) + urllib.parse.urlsplit(url)[1:])


class RecentLoginRequiredMixin(LoginRequiredMixin):
    """
    Require the user to have logged in within a number of seconds.
    """

    max_last_login_delta = 1800  # Defaults to 30 minutes

    def dispatch(self, request, *args, **kwargs):
        """Call the appropriate method if the user's login is recent"""
        resp = super().dispatch(request, *args, **kwargs)

        if resp.status_code == 200:
            delta = datetime.timedelta(seconds=self.max_last_login_delta)
            if now() > (request.user.last_login + delta):
                return logout_then_login(request, self.get_login_url())
        return resp



import os
import io
import secrets
import random
from typing import Any, Set, List, TypeVar, Type
from datetime import datetime, timezone
from random import randint

from django.db.models import Model
from django.utils.text import slugify

try:
    from PIL import Image, ImageDraw, ImageFont
except ImportError:
    raise ImportError(  # noqa: B904
        "Install `pillow` package. Run `pip install pillow`."
    )

try:
    from mimesis.locales import Locale
    from mimesis import (
        Text,
        Person,
        Numeric,
        Finance,
        Datetime,
        Address,
        Food,
        Internet,
    )
except ImportError:
    raise ImportError(  # noqa: B904
        "Install `mimesis` package. Run `pip install mimesis`."
    )

T = TypeVar("T", bound=Model)


class BaseDataGenerator:
    """Generate reusable data"""

    def __init__(self, locale="en"):
        self.person = Person(getattr(Locale, locale.upper()))
        self.text = Text(getattr(Locale, locale.upper()))
        self.fiance = Finance(getattr(Locale, locale.upper()))
        self.address = Address(getattr(Locale, locale.upper()))
        # self.internet = Internet(getattr(Locale, locale.upper()))

    def create_placeholder_image(
        self,
        number: int,
        subject: str,
        size: tuple[int, int] = (440, 660),
        pic_format: str = "WEBP",
    ):
        """
        Create a gray placeholder image with a number in a specified format.
        Default format is WEBP.
        """
        SUPPORTED_FORMATS = ["JPEG", "PNG", "WEBP"]
        desired_format = pic_format.upper()
        if desired_format not in SUPPORTED_FORMATS:
            raise ValueError(
                f"Unsupported format: {desired_format}. Supported formats are: {', '.join(SUPPORTED_FORMATS)}"
            )

        light_gray_color = (216, 216, 216)
        image = Image.new("RGB", size, color=light_gray_color)
        draw = ImageDraw.Draw(image)
        font_size = 20  # Adjust as needed
        font = ImageFont.load_default()

        # Prepare the text
        number_text = f"Pic {number}"
        dimension_text = f"{size[0]} x {size[1]}"

        # Manual estimation of text size
        approx_char_width = font_size * 0.6  # Approximation for character width
        number_text_width = approx_char_width * len(number_text)
        dimension_text_width = approx_char_width * len(dimension_text)
        text_height = font_size  # Approximation for text height

        # Calculate positions
        number_text_x = (size[0] - number_text_width) / 2
        number_text_y = (size[1] - text_height) / 2 - 10

        dimension_text_x = (size[0] - dimension_text_width) / 2
        dimension_text_y = number_text_y + text_height + 5

        # Draw the texts
        draw.text((number_text_x, number_text_y), number_text, font=font, fill="black")
        draw.text(
            (dimension_text_x, dimension_text_y),
            dimension_text,
            font=font,
            fill="black",
        )

        # Save the image to a bytes object
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format=desired_format)
        img_byte_arr = img_byte_arr.getvalue()

        width, height = size
        filename = f"{subject}_{number}_{width}x{height}.{desired_format.lower()}"

        return img_byte_arr, filename, desired_format

    def get_random_secret(self, nbytes=20):
        """Generate random string"""
        return secrets.token_urlsafe(nbytes)

    def get_random_words(self, quantity):
        return " ".join(self.text.words(quantity=quantity))

    def get_random_color(self):
        """Generate random code"""
        return Text().color()

    def get_random_int(self, start: int, end: int) -> int:
        return random.randint(start, end)

    def get_random_hex_code(self):
        """Generate random hex code"""
        return Text().hex_color()

    def get_random_time(self):
        """Generate random time"""
        return Datetime().time()

    def get_random_datetime(self, start=None, end=None):
        """Generate a random time between start and end."""
        if start is None:
            current_year = timezone.now().year
            start = timezone.make_aware(datetime(current_year, 1, 1))

        if end is None:
            current_year = timezone.now().year
            end = timezone.make_aware(datetime(current_year, 12, 31))

        start_timestamp = start.timestamp()
        end_timestamp = end.timestamp()
        random_timestamp = random.uniform(start_timestamp, end_timestamp)
        random_datetime = timezone.make_aware(datetime.fromtimestamp(random_timestamp))
        return random_datetime

    def get_random_boolean(self):
        """Generate a random boolean."""
        return random.choice([True, False])

    def get_random_booleans(self, total_true=1, total_false=1, is_shuffle=False):
        trues = [True] * total_true
        falses = [False] * total_false
        total_booleans = trues + falses
        shuffle_booleans = total_booleans[:]
        random.shuffle(shuffle_booleans)
        booleans = shuffle_booleans if is_shuffle else total_booleans
        return booleans

    def get_random_population(self, population, k):
        objs = set(random.sample(population, k))
        population = set(population)
        return list(objs), list(population - objs)

    def get_random_object(self, population):
        return random.choice(population)

    def get_random_float(self, lower, upper):
        """Generate a random float with 2 digits after seperator"""
        return round(random.uniform(lower, upper), 2)

    def get_random_currency(self):
        """Generate a random currency."""
        return random.choice(["IRR", "USD", "EUR"])

    def get_random_price(self):
        """Generate a random price."""
        return Finance().price()

    def get_random_number(self, start=-100, stop=1000):
        """Generate a random number."""
        return Numeric().integer_number(start=start, end=stop)

    def get_random_County_code(self):
        """Generate a random County code"""
        return Address().country_code()

    def get_random_city(self):
        """Generate a random city."""
        return Address().city()

    def get_random_address(self):
        """Generate a random address."""
        return Address().address()

    def get_random_postal_code(self):
        """Generate a random postal_code."""
        return Address().postal_code()

    def get_random_street_number(self):
        """Generate a random street_number."""
        return Address().street_number()

    def get_random_full_name(self):
        """Generate a random full_name."""
        return Person().full_name()

    def get_random_first_name(self):
        return Person().first_name()

    def get_random_last_name(self):
        return Person().last_name()

    def get_random_telephone(self):
        """Generate a random telephone."""
        return slugify(Person().telephone())

    def get_random_status(self):
        """Generate a random status."""
        status = [
            "waiting",
            "expiring",
            "cancelled",
            "shipped",
            "processing",
            "delivered",
            "completed",
        ]
        return random.choice(status)

    def get_image_banner(self, index):
        """Generate a random image banner."""
        if index == 0:
            result = "first"
        elif index == 1:
            result = "second"
        else:
            result = "other"
        return result

    def get_random_image(self, order):
        """Generate a random image choice."""
        if order == 1:
            result = "first"
        elif order == 2:
            result = "second"
        else:
            result = "other"
        return result

    def get_random_sentence(self):
        return Text().sentence()

    def get_random_email(self):
        return Person().email()

    def get_random_job(self):
        return Person().occupation()

    def get_random_gender(self):
        """Generate a random gender."""
        gender = ["male", "female"]
        return random.choice(gender)

    def get_random_province(self):
        return Address().province()

    def get_unique_phone_number_set(self, min_length: int, digits: int) -> Set[str]:
        """
        generates a unique set of phone number which is guaranteed to be bigger
        than a specified number and it contains given digits or characters given.

        PARAMS
        ------
        min_length: int:
            the minimum element_length of the resulting set
        digits: int:
            how many characters or digits should each phone number have.

        Returns
        -------
            a set containing unique phone numbers, its element_length is always
            somewhere between min_length and (min_length + 10)
        """
        phone_number_set = set(
            [self.get_random_telephone()[:digits] for _ in range(min_length)]
        )
        while len(phone_number_set) < min_length:
            phone_number_set = set.union(
                phone_number_set,
                set([self.get_random_telephone()[:digits] for _ in range(10)]),
            )
        return phone_number_set

    def get_currency_exchange(
        self, source_currency, source_value, rial_profit, toman_profit, usd_profit
    ):
        if source_currency == "R":
            target_value = int(int(source_value) * random.uniform(*rial_profit))
        elif source_currency == "T":
            target_value = int(int(source_value) * random.uniform(*toman_profit))
        else:
            target_value = round(int(source_value) * random.uniform(*usd_profit), 2)
        return target_value

    def get_voucher_kind(self, static_chance: int = 20):
        """
        Get a random kind of voucher.

        PARAMS:
        -----
        static_chance: int = 20:
            chance for every created voucher to be static.
        """

        kinds = ("static_based", "code_based")
        return random.choices(
            kinds, weights=(static_chance, (100 - static_chance)), k=1
        )[0]

    def get_voucher_type(self):
        """
        Get a random type for voucher.
        """
        types = ("fixed_price_based", "percentage_based")
        return random.choice(types)

    def get_voucher_status(self):
        """
        Get a random status for voucher.
        """
        status = ("open", "suspend", "consumed")
        return random.choice(status)

    def get_random_percentage(self, start: int = 1, end: int = 100):
        """
        Get a random percentage between given start and end numbers.
        raises a value error if given values are bellow0 or over 100 or if start
        is bigger than end.
        """
        if 100 < start < 0:
            raise ValueError(
                f"start has to be between 0 to 100" f",however given start is {start}"
            )
        elif 100 < end < 0:
            raise ValueError(
                f"end has to be between 0 to 100" f",however given start is {end}"
            )
        elif start > end:
            raise ValueError(
                f"start has to be smaller than end, however given"
                f"start is {start} and given end is {end}"
            )
        else:
            return random.randint(start, end)

    def get_unique_hashes_list(self, total: int, element_length: int = 12):
        """
        Get a set(which has unique elements) of hashes.

        PARAMS:
        total: int:
            length of the set.
        element_length: int = 12:
            length of each element in the set.
        """
        hash_set = {self.get_random_secret(element_length) for _ in range(total)}
        while len(hash_set) < total:
            hash_set = set.union(
                hash_set, {self.get_random_secret(element_length) for _ in range(10)}
            )
        return list(hash_set)[:total]

    def get_random_time_between_two_datetime_objects(
        self, time1: datetime, time2: datetime, tz: timezone = timezone.utc
    ):
        """
        Get a random datetime object between two given datetime objects.
        raises ValueError if start time is bigger than end time.

        PARAMS
        -----
        time1: datetime:
            start time.
        time2: datetime:
            end time.
        tz: timezone = timezone.utc:
            requested time zone, default is utc.
        """
        time1 = int(datetime.timestamp(time1))
        time2 = int(datetime.timestamp(time2))
        if time1 > time2:
            raise ValueError("start time is later than end time" f"{time1} > {time2}")
        random_time = random.randint(time1, time2)
        return datetime.fromtimestamp(random_time).replace(tzinfo=tz)

    def get_random_spice(self):
        """
        Gets the name of a random spice.
        """
        return Food().spices()

    def add_to_m2m(
        self, objs: List[Any], target_field: str, item_pre_obj: int, item: Type[T]
    ) -> None:
        """Add a number of randomly selected objects from a list to a ManyToManyField.

        This method selects `item_pre_obj` number of objects randomly from the `objs` list and adds them to the ManyToManyField
        on the model instance. The `objs` list must not be empty, or an `IndexError` will be raised.

        Args:
            self: An instance of the model.
            objs (List[Any]): A list of objects to select from.
            target_field (str): The name of the ManyToManyField on the model instance.
            item_pre_obj (int): The number of objects to add to the ManyToManyField.
            item (Type[T]): The type of the model that the ManyToManyField belongs to.

        Raises:
            IndexError: If the `objs` list is empty.

        Returns:
            None.
        """

        attr = getattr(item, target_field)
        try:
            items_to_add = list(map(lambda _: random.choice(objs), range(item_pre_obj)))
        except IndexError as e:
            raise IndexError("`objs` is empty. Please ensure population is not None.")
        attr.add(*items_to_add)



from typing import Any

from django.conf import settings
from django.utils.text import slugify


class SlugService:
    """
    A service class for handling slug creation and uniqueness for a given model instance.

    The `SlugService` class provides methods to create slugs from the instance title,
    check if a slug has been modified, and ensure the uniqueness of slugs within the model.
    It uses the `AUTO_SLUGIFY_ENABLED` setting to determine whether to automatically generate
    slugs from the instance title.
    """

    def __init__(self, instance: Any) -> None:
        self.instance = instance
        self.auto_slugify_enabled: bool = getattr(
            settings, "AUTO_SLUGIFY_ENABLED", True
        )

    def _create_slug(self) -> str:
        """Generate a slug from the instance title if auto-slugify is enabled, otherwise use the existing slug."""
        if self.auto_slugify_enabled:
            return slugify(self.instance.title, allow_unicode=True)
        return self.instance.slug

    def _is_slug_unique(self, slug: str) -> bool:
        """Check if a given slug is unique among the instances."""
        return (
            not type(self.instance)
            .objects.filter(slug=slug)
            .exclude(pk=self.instance.pk)
            .exists()
        )

    def _generate_unique_slug(self, base_slug: str) -> str:
        """Generate a unique slug by appending a counter if needed."""
        new_slug = base_slug
        counter = 1
        while not self._is_slug_unique(new_slug):
            new_slug = f"{base_slug}-{counter}"
            counter += 1
        return new_slug

    def has_slug_changed(self, new_slug: str) -> bool:
        """Check if the slug has been modified compared to the stored slug."""
        if not self.instance.pk:
            return False
        existing_instance = type(self.instance).objects.get(pk=self.instance.pk)
        return existing_instance.slug != new_slug

    def create_unique_slug(self) -> str:
        """Create and return a unique slug for the instance."""
        base_slug = self._create_slug()
        return self._generate_unique_slug(base_slug)



from .http import HeaderMixin
from .locale import SetLanguageMixinView
from.access import (
    AccessMixin,
    LoginRequiredMixin,
    AnonymousRequiredMixin,
    PermissionRequiredMixin,
    MultiplePermissionsRequiredMixin,
    GroupRequiredMixin,
    UserPassesTestMixin,
    SuperuserRequiredMixin,
    StaffuserRequiredMixin,
    SSLRequiredMixin,
    RecentLoginRequiredMixin,
)
from .cache import (
    CacheControlMixin,
    NeverCacheMixin,
)

__all__ = [
    "HeaderMixin",
    "SetLanguageMixinView",
    "AccessMixin",
    "LoginRequiredMixin",
    "AnonymousRequiredMixin",
    "PermissionRequiredMixin",
    "MultiplePermissionsRequiredMixin",
    "GroupRequiredMixin",
    "UserPassesTestMixin",
    "SuperuserRequiredMixin",
    "StaffuserRequiredMixin",
    "SSLRequiredMixin",
    "RecentLoginRequiredMixin",
    "CacheControlMixin",
    "NeverCacheMixin",
]



from django.conf import settings
from django.http import HttpRequest


def get_maintenance_variables(request: HttpRequest):
    context = dict()
    context["is_maintenance_enabled"] = getattr(settings, "UNDER_CONSTRUCTION_MODE")
    context["is_coming_soon_enabled"] = getattr(settings, "COMING_SOON_MODE")
    return context



"""
ForeignKeyLinker Module

This module contains a class, ForeignKeyLinker, designed to generate
a function for linking to a related object's change page in the Django admin interface.
The generated function can be used in a Django ModelAdmin class within the list_display
property to display clickable links to related objects.

Classes:
    - ForeignKeyLinker: Generates link functions for related objects in Django admin.

Attributes:
    - field_name (str): The name of the foreign key field in the model.
    - short_description (str, optional): A short description for the function,
    used as the column header. Defaults to the capitalized field name with
    underscores replaced by spaces.
"""
from django.urls import reverse
from django.utils.html import format_html


class ForeignKeyLinker:
    """
    This class is designed to generate a function for linking to a related object's
    change page in the Django admin interface.

    The function generated by an instance of this class can be used in a Django
    ModelAdmin class within the list_display property.

    Attributes:
        field_name (str): The name of the foreign key field in the model.
        short_description (str, optional): The short description for the function.
            If not provided, it defaults to the capitalized field name with underscores
            replaced by spaces.
    """

    def __init__(self, field_name, short_description=None):
        """
        Initialize ForeignKeyLinker with field_name and an optional short_description.
        """
        self.field_name = field_name
        self.short_description = (
            short_description or field_name.replace("_", " ").capitalize()
        )

    def link_to_fk(self, obj):
        """
        Return an HTML link to the related object's change page in the
        Django admin interface.

        Args:
            obj (Model instance): The instance of the model in which the
            foreign key is defined.

        Returns:
            str: An HTML string representing a link to the related object's change page,
            or None if the related object does not exist.
        """
        # Get the related object
        related_obj = getattr(obj, self.field_name)

        if related_obj is None:
            return None

        # Generate the url to the related object's change page
        url = reverse(
            f"admin:{related_obj._meta.app_label}_{related_obj._meta.model_name}_change",
            args=[related_obj.pk],
        )

        # Return the HTML string
        return format_html('<a href="{}">{}</a>', url, related_obj)

    def get_link(self):
        """
        Return the function for use in a Django ModelAdmin class.

        Returns:
            Function: The function to be used in a Django ModelAdmin class.
        """

        # Define the function
        def func(obj):
            return self.link_to_fk(obj)

        func.__name__ = self.field_name
        func.short_description = self.short_description
        func.admin_order_field = self.field_name

        # Return the function
        return func



"""
This module provides a set of functions to convert between different
units of measurement.
It includes conversions between bytes and megabytes, days and seconds,
and seconds and minutes.
"""

from sage_tools.helpers.typings import Byte, Day, MegaByte, Minute, Second


class UnitConvertor:
    """
    A utility class for converting units between various scales of measurement.

    This class provides static methods for converting between bytes and megabytes,
    days and seconds, and seconds and minutes.
    """

    @staticmethod
    def convert_byte_to_megabyte(value: Byte) -> MegaByte:
        """
        Convert bytes to megabytes.

        :param value: The number of bytes.
        :type value: Byte
        :return: The equivalent number of megabytes.
        :rtype: MegaByte
        """
        return value * 1e-6

    @staticmethod
    def convert_megabyte_to_byte(value: MegaByte) -> Byte:
        """
        Convert megabytes to bytes.

        :param value: The number of megabytes.
        :type value: MegaByte
        :return: The equivalent number of bytes.
        :rtype: Byte
        """
        return value * 1_000_000

    @staticmethod
    def convert_days_to_seconds(value: Day) -> Second:
        """
        Convert days to seconds.

        :param value: The number of days.
        :type value: Day
        :return: The equivalent number of seconds.
        :rtype: Second
        """
        return 60 * 60 * 24 * value

    @staticmethod
    def convert_seconds_to_minutes(value: Second) -> Minute:
        """
        Convert seconds to minutes.

        :param value: The number of seconds.
        :type value: Second
        :return: The equivalent number of minutes.
        :rtype: Minute
        """
        return value / 60.0

    @staticmethod
    def convert_minutes_to_seconds(value: Minute) -> Second:
        """
        Convert minutes to seconds.

        :param value: The number of minutes.
        :type value: Minute
        :return: The equivalent number of seconds.
        :rtype: Second
        """
        return int(value * 60)



import logging
from datetime import timedelta
from typing import Any, Optional

from django.conf import settings
from django.http import HttpRequest
from django.utils import timezone

try:
    from cryptography.fernet import InvalidToken
except ImportError:
    raise ImportError("Install `cryptography` package. Run `pip install cryptography`.")

from sage_tools.encryptors import FernetEncryptor

logger = logging.getLogger(__name__)


class SessionHandler:
    """
    Manages session variables with encryption and a custom expiry time for enhanced security and privacy.

    This class provides methods to set, get, and delete session variables securely. It uses
    Fernet symmetric encryption to ensure that session data is encrypted before being stored,
    offering an additional layer of security. Each session variable can have its own lifespan,
    after which it is considered expired and automatically removed.
    """

    def __init__(self, request: HttpRequest) -> None:
        """
        Initializes the SessionHandler with the current request and uses the secret key from Django settings for encryption.
        """
        self.request = request
        self.fernet = FernetEncryptor(settings.FERNET_SECRET_KEY)

    def set(
        self, key: str, value: str, lifespan=timedelta(minutes=10), encrypt=True
    ) -> None:
        """
        Encrypts and sets a session variable with a specified lifespan.
        """
        if not isinstance(key, str) or not key:
            raise ValueError("Key must be a non-empty string")
        if not isinstance(lifespan, timedelta) or lifespan.total_seconds() <= 0:
            raise ValueError("Lifespan must be a positive timedelta object")

        try:
            encrypted_value = (
                self.fernet.encrypt(value.encode("utf-8")) if encrypt else value
            )
            self.request.session[key] = {
                "value": encrypted_value,
                "created_at": timezone.now().timestamp(),
                "lifespan": lifespan.total_seconds(),
            }
        except Exception as e:
            logger.error(f"Error encrypting session data for key {key}: {str(e)}")

    def get(self, key: str, decrypt=True) -> Optional[str]:
        """
        Retrieves, decrypts, and returns the value of a session variable if it has not expired.
        """
        session_info = self.request.session.get(key)
        if session_info and self._is_valid_session_data(session_info):
            created_at = session_info.get("created_at")
            expiry = session_info.get("lifespan", 0)
            if timezone.now().timestamp() - created_at < expiry:
                encrypted_value = session_info.get("value")
                try:
                    return (
                        self.fernet.decrypt(encrypted_value)
                        if decrypt
                        else encrypted_value
                    )
                except InvalidToken:
                    logger.error(
                        f"Invalid token for session key {key}. Possible data tampering."
                    )
            else:
                self.delete(key)
        return None

    def delete(self, key: str) -> Optional[Any]:
        """
        Deletes a session variable, if it exists.
        """
        return self.request.session.pop(key, None)

    def is_expired(self, key: str) -> bool:
        """
        Checks if a session variable has expired.
        """
        session_info = self.request.session.get(key)
        if session_info and self._is_valid_session_data(session_info):
            created_at = session_info.get("created_at")
            lifespan = session_info.get("lifespan", 0)
            return timezone.now().timestamp() - created_at >= lifespan
        return True

    def refresh(self, key: str, lifespan=timedelta(minutes=10)) -> bool:
        """
        Refreshes the lifespan of an existing session variable, if it exists and has not expired.
        """
        if not self.is_expired(key):
            session_info = self.request.session.get(key)
            if session_info:
                session_info["created_at"] = timezone.now().timestamp()
                session_info["lifespan"] = lifespan.total_seconds()
                self.request.session[key] = session_info
                return True
        return False

    def exists(self, key: str) -> bool:
        """
        Checks if a session variable exists and has not expired.
        """
        return key in self.request.session and not self.is_expired(key)

    def _is_valid_session_data(self, session_data: dict[str, Any]) -> bool:
        """
        Validates the structure of the session data.
        """
        return all(k in session_data for k in ["value", "created_at", "lifespan"])

    def flush(self) -> None:
        """
        Clears the session data and regenerates a new session key.
        """
        self.request.session.flush()

    def cycle_key(self) -> None:
        """
        Preserves the session data but changes the session key to prevent session fixation.
        """
        self.request.session.cycle_key()



"""
This module provides middleware support for handling user language
preferences in a Django web application. 
It extends Django's built-in locale middleware functionality,
offering a more nuanced approach to 
language detection and selection through URL prefixes and cookies.
The core component, CookieLocaleMiddleware, 
intercepts web requests to determine the most suitable language
for the user and adjust the request or response accordingly.
"""

import logging

from django.conf import settings
from django.middleware.locale import LocaleMiddleware as DjangoLocaleMiddleware
from django.shortcuts import redirect
from django.utils import translation

from sage_tools.utils.locale import MultilingualService

logger = logging.getLogger(__name__)


class CookieLocaleMiddleware(DjangoLocaleMiddleware):
    """
    Extends Django's LocaleMiddleware to manage language settings more dynamically.

    The middleware determines the user's preferred language by examining the URL and cookies,
    then aligns the request with the determined preference. It's designed to work seamlessly
    with multilingual sites, redirecting users to the appropriate language version of the site
    based on their preferences and the available languages.
    """

    def process_request(self, request):
        """
        Examines and aligns the request with the user's preferred language based
        on the URL or cookies.

        This method checks the request path for language prefixes (e.g., '/en/', '/fr/')
        and compares it with the language specified in the user's language cookie.
        It activates the appropriate language for the session and redirects to a URL
        that matches the preferred language if necessary.

        Example:
            When a user accesses a URL like '/es/about/', but their language cookie
            is set to 'en', this method will redirect them to '/en/about/' if
            English is an available language. Conversely, if the URL is '/about/' with no
            language prefix and the cookie is set to 'es', the user will be redirected to '/es/about/'.

        Parameters:
            request: HttpRequest object containing metadata about the request.

        Returns:
            A redirect response if the language from the URL and cookie differ, otherwise None.
        """

        # Exclude certain paths from language prefix redirection
        if request.path_info.startswith(
            "/set-language/"
        ) or request.path_info.startswith("/i18n/"):
            return super().process_request(request)

        # Retrieve the language from the URL and the cookie
        url_language = translation.get_language_from_path(request.path_info)
        cookie_language = request.COOKIES.get(settings.LANGUAGE_COOKIE_NAME)

        # Ensure that the URL is clean of language prefixes for accurate redirection
        clean_path = MultilingualService.remove_language_prefix(request.path_info)

        # Check if the cookie language is valid
        if cookie_language in dict(settings.LANGUAGES):
            # Activate the cookie language and set it in the request
            translation.activate(cookie_language)
            request.LANGUAGE_CODE = cookie_language

            # Redirect to the correct language URL, considering the default language case
            if cookie_language != settings.LANGUAGE_CODE and (
                not url_language or url_language != cookie_language
            ):
                new_path = MultilingualService.add_language_prefix(
                    clean_path, cookie_language
                )
                return redirect(new_path)
            elif (
                cookie_language == settings.LANGUAGE_CODE
                and url_language
                and url_language != settings.LANGUAGE_CODE
            ):
                # Redirect to the clean path if the default language is set in the cookie and
                #  URL has a non-default prefix
                return redirect(clean_path)

        # If the URL language is valid, activate it
        elif url_language in dict(settings.LANGUAGES):
            translation.activate(url_language)
            request.LANGUAGE_CODE = url_language
            # Redirect if the URL language is different from the cookie language and is not
            #  the default language
            if (
                url_language != settings.LANGUAGE_CODE
                and url_language != cookie_language
            ):
                new_path = MultilingualService.add_language_prefix(
                    clean_path, url_language
                )
                return redirect(new_path)

        # Handle default language without prefix
        elif not url_language or url_language == settings.LANGUAGE_CODE:
            translation.activate(settings.LANGUAGE_CODE)
            request.LANGUAGE_CODE = settings.LANGUAGE_CODE
            # Redirect only if there was a language prefix and the cookie language is not
            #  the default language
            if url_language and cookie_language != settings.LANGUAGE_CODE:
                return redirect(clean_path)

        # Fallback to default language handling
        super().process_request(request)

    def process_response(self, request, response):
        """
        Modifies the response to set the correct language cookie based on
        the user's preferred language.

        After the parent class's process_response method runs, this method
        checks if the current language in the session differs from the one
        in the user's language cookie. If there is a difference, it updates
        the language cookie to the current session language.

        Example:
            If a user's session language is 'fr' (French) after interacting
            with the site, but their language cookie is still 'en' (English),
            this method will set the cookie to 'fr' in the response.
            The next time the user makes a request, their preferred
            language will be French.

        Parameters:
            request: HttpRequest object containing metadata about the request.
            response: HttpResponse object that will be sent back to the user.

        Returns:
            The modified HttpResponse object with the updated language cookie if necessary.
        """

        # First, call the parent class's process_response method
        response = super().process_response(request, response)

        # Get the current language
        current_language = translation.get_language()

        # Set the language cookie if it's different from the current one
        if (
            current_language
            and request.COOKIES.get(settings.LANGUAGE_COOKIE_NAME) != current_language
        ):
            response.set_cookie(settings.LANGUAGE_COOKIE_NAME, current_language)

        return response



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from sage_tools.validators.numeral import HalfPointIncrementValidator


class RatingMixin(models.Model):
    """
    An abstract base class model designed to add a rating functionality to any model that requires it.
    This mixin introduces a standardized way to capture user ratings for various objects within the
    application, ensuring consistency and reusability across different parts of the system.

    The rating system implemented by this mixin allows for granular feedback, supporting not only whole
    numbers but also half-point increments within a defined range from 1 to 5. This flexibility accommodates
    a wide range of rating scenarios, from product reviews to service evaluations, enhancing the application's
    ability to gather and reflect nuanced user feedback.

    As an abstract model, RatingMixin does not create a database table on its own but serves as a building
    block for other models, allowing them to easily integrate a uniform rating system without the need for
    redundant code.
    """

    rating = models.FloatField(
        verbose_name=_("Rating"),
        default=1,
        help_text=_("Rate from 1 (lowest) to 5 (highest)."),
        validators=[
            MinValueValidator(1, message=_("Rating must be at least 1.")),
            MaxValueValidator(5, message=_("Rating cannot be more than 5.")),
            HalfPointIncrementValidator(
                message=_("Rating must be between 1 and 5 in half-point increments.")
            ),
        ],
        db_comment="User's rating from 1 to 5.",
    )

    class Meta:
        abstract = True



from .base import Encryptor

try:
    from cryptography.fernet import Fernet, InvalidToken
except ImportError:
    raise ImportError("Install `cryptography` package. Run `pip install cryptography`.")


class FernetEncryptor(Encryptor):
    """
    Fernet encryption class.

    This class uses the Fernet symmetric encryption method provided by the `cryptography` package to encrypt and decrypt data.

    Parameters
    ----------
    secret_key : str
        The secret key used for encryption and decryption. Must be a URL-safe base64-encoded 32-byte key.

    Methods
    -------
    encrypt(data: str) -> str
        Encrypts the given data using Fernet encryption.
    decrypt(data: str) -> str
        Decrypts the given data using Fernet encryption.

    Examples
    --------
    >>> secret_key = Fernet.generate_key()
    >>> encryptor = FernetEncryptor(secret_key)
    >>> encrypted_data = encryptor.encrypt("Hello, World!")
    >>> decrypted_data = encryptor.decrypt(encrypted_data)
    >>> decrypted_data
    'Hello, World!'
    """

    def __init__(self, secret_key: str):
        self.fernet = Fernet(secret_key)

    def encrypt(self, data: str) -> str:
        """
        Encrypts the given data using Fernet encryption.

        Parameters
        ----------
        data : str
            The data to be encrypted.

        Returns
        -------
        str
            The encrypted data.
        """
        self._validate_data(data)
        data = self._encode_data(data)
        encrypted_value = self.fernet.encrypt(data)
        return encrypted_value.decode("utf-8")

    def decrypt(self, data: str) -> str:
        """
        Decrypts the given data using Fernet encryption.

        Parameters
        ----------
        data : str
            The data to be decrypted.

        Returns
        -------
        str
            The decrypted data.
        """
        self._validate_data(data)
        data = self._encode_data(data)
        return self.fernet.decrypt(data).decode("utf-8")

    def _validate_data(self, data):
        """
        Validates the data to ensure it is either a string or bytes.

        Parameters
        ----------
        data : str or bytes
            The data to be validated.

        Raises
        ------
        TypeError
            If the data is not a string or bytes.
        """
        if not isinstance(data, (str, bytes)):
            raise TypeError(
                "FernetEncryptor only supports string or bytes data types for encryption."
            )

    def _encode_data(self, data):
        """
        Encodes the data to bytes if it is a string.

        Parameters
        ----------
        data : str or bytes
            The data to be encoded.

        Returns
        -------
        bytes
            The encoded data.
        """
        if isinstance(data, str):
            data = data.encode("utf-8")
        return data



from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _
import re

@deconstructible
class NameValidator:
    message = _(
        "Enter a valid name. The name can only contain letters, spaces, hyphens, and apostrophes. Examples: John Doe, Mary-Jane O'Connor."
    )
    code = "invalid_name"
    regex = r"^[a-zA-Z]+([ '\-][a-zA-Z]+)*$"

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        if not re.match(self.regex, value):
            raise ValidationError(self.message, code=self.code)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.message == other.message
            and self.code == other.code
            and self.regex == other.regex
        )



from django import forms

try:
    import pytz
except ImportError:
    raise ImportError(  # noqa: B904
        "Install `pytz` package. Run `pip install pytz`."
    )

class TimezoneForm(forms.Form):
    timezone = forms.ChoiceField(choices=[(tz, tz) for tz in pytz.all_timezones])
    next = forms.CharField(widget=forms.HiddenInput())



class HeaderMixin:
    """
    A mixin to add extra HTTP headers to the response in Django views.

    This mixin allows specifying additional HTTP headers that will be included
    in the response generated by the view. Headers can be defined statically
    by setting the `headers` attribute or dynamically by overriding the
    `get_headers` method.
    """

    headers = {}

    def get_headers(self, request):
        return self.headers

    def dispatch(self, request, *args, **kwargs):
        """
        Overrides the dispatch method to add the specified headers to the response
        before returning it.
        """
        response = super().dispatch(request, *args, **kwargs)
        for key, value in self.get_headers(request).items():
            if key not in response:
                response[key] = value
        return response



"""
This module provides a Django model class for image manipulation and retrieval.
It includes methods for converting image formats, cropping images, and generating
thumbnails.
"""

import io
import os
from pathlib import Path

from django.core.files import File
from django.core.validators import MaxLengthValidator, MinLengthValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

try:
    from PIL import Image
except ImportError as e:
    raise ImportError("Install `pillow` package. Run `pip install pillow`.")

try:
    from sorl.thumbnail import get_thumbnail
except ImportError as e:
    raise ImportError(
        "Install `sorl-thumbnail` package. Run `pip install sorl-thumbnail`."
    )


class PictureOperationAbstract(models.Model):
    """
    An abstract Django model class for performing various operations on images.
    This class provides functionalities for image conversion, cropping, thumbnail
    generation, and retrieving image properties like URL, size, and dimensions.
    """

    picture: str

    alternate_text = models.CharField(
        _("Picture Alternate Text"),
        max_length=110,
        validators=[MaxLengthValidator(150), MinLengthValidator(3)],
        null=True,
        blank=True,
        help_text=_("Write about picture for SEO"),
    )

    width_field = models.PositiveSmallIntegerField(
        _("Picture Width"),
        null=True,
        blank=True,
        editable=False,
        help_text=_("size of picture's Width"),
    )

    height_field = models.PositiveSmallIntegerField(
        _("Picture Height"),
        null=True,
        blank=True,
        editable=False,
        help_text=_("size of picture's Height"),
    )

    class Meta:
        """
        Meta
        """

        abstract = True

    def convert_to_webp(self):
        """
        Converts the image associated with this model instance to the WEBP format.
        The converted image replaces the original image.
        """
        if ".webp" not in self.picture.path:
            path = self.picture.path
            # format_ = Path(path).suffix
            with Image.open(path) as data:
                data_rgb = data.convert("RGB")
                output = io.BytesIO()
                data_rgb.save(output, format="webp")
                self.picture.save(
                    os.path.basename(path).replace(format, ".webp"), output
                )

    def convert_to_jpg(self):
        """
        Converts the image associated with this model instance to JPEG format.
        The converted image replaces the original image.
        """
        if ".jpg" not in self.picture.path:
            path = self.picture.path
            format = Path(path).suffix

            with Image.open(path) as data:
                data_rgb = data.convert("RGB")
                output = io.BytesIO()
                data_rgb.save(output, format="JPEG")
                self.picture.save(
                    os.path.basename(path).replace(format, ".jpg"), output
                )

    def convert_to_png(self):
        """
        Converts the image associated with this model instance to PNG format.
        The converted image replaces the original image and the original file is removed.
        """
        if ".png" not in self.picture.path:
            path = self.picture.path
            format = Path(path).suffix

            with File(open(self.picture.path, "rb")) as data:
                self.picture.save(
                    os.path.basename(self.picture.path).replace(format, ".png"), data
                )
        try:
            os.remove(path)
        except FileNotFoundError:
            pass

    def crop(self, left, top, right, bottom):
        """
        Crops the image associated with this model instance.

        Parameters:
            left (int): The left boundary of the crop box.
            top (int): The upper boundary of the crop box.
            right (int): The right boundary of the crop box.
            bottom (int): The lower boundary of the crop box.
        """
        path = self.picture.path
        with Image.open(path) as pic:
            pic_rgb = pic.convert("RGB")
            data = pic_rgb.crop((left, top, right, bottom))
            output = io.BytesIO()
            data.save(output, format="JPEG")
            self.picture.save(os.path.basename(path), output)

    def get_thumbnail(self, size="100x100"):
        """
        Generates a thumbnail for the image associated with this model instance.
        """
        return get_thumbnail(self.picture, crop="center", quality=99)

    def get_thumbnail_url(self, size="100x100"):
        """
        Retrieves the URL of the thumbnail for the image associated with this model instance.
        """
        return get_thumbnail(self.picture, crop="center", quality=99).url

    def get_picture_url(self):
        """
        Retrieves the URL of the image associated with this model instance.
        """
        return self.picture.url

    def get_picture_size(self):
        """
        Retrieves the file size of the image associated with this model instance.
        """
        return self.picture.size

    def get_picture_dimensions(self):
        """
        Retrieves the dimensions of the image associated with this model instance.
        """
        return (self.picture.width, self.picture.height)

    def get_file_name(self):
        """
        Retrieves the file name of the image associated with this model instance.
        """
        return os.path.basename(self.picture.name)



from django.apps import apps
from django.contrib import admin
from django.contrib.admin.sites import site


def get_app_list(self, request, app_label=None):
    """
    Return a sorted list of all the installed apps that have been
    registered in this site.
    """
    app_dict = self._build_app_dict(request)

    # Filter by app_label if provided
    if app_label:
        app_dict = {app_label: app_dict[app_label]} if app_label in app_dict else {}

    if not app_dict:
        return

    for app_name in list(app_dict.keys()):
        app = app_dict[app_name]
        model_priority = {}
        for model in app["models"]:
            try:
                model_instance = apps.get_model(app_name, model["object_name"])
                model_priority[model["object_name"]] = getattr(
                    site._registry[model_instance],
                    "admin_priority",
                    20,
                )
            except LookupError:
                # This handles the case where a model cannot be found (e.g., for the 'constance' app)
                continue

        # Sort models in the app by their priority
        app["models"].sort(key=lambda x: model_priority.get(x["object_name"], 20))

    # Sort the apps by their name
    app_list = sorted(app_dict.values(), key=lambda x: x["name"].lower())
    return app_list


admin.AdminSite.get_app_list = get_app_list



from .user import (
    UserFormKwargsMixin,
    MessageMixin,
    FormValidMessageMixin,
    FormInvalidMessageMixin,
    FormMessagesMixin,
    UserKwargModelFormMixin,
)

__all__ = [
    "UserFormKwargsMixin",
    "MessageMixin",
    "FormValidMessageMixin",
    "FormInvalidMessageMixin",
    "FormMessagesMixin",
    "UserKwargModelFormMixin",
]



from .base import (
    TimeStampMixin,
    UUIDBaseModel,
    StockUnitMixin,
    BaseTitleSlugMixin,
    TitleSlugMixin,
    BaseTitleSlugDescriptionMixin,
    TitleSlugDescriptionMixin,
)

from .address import AddressMixin
from .access import SingletonModelMixin
from .rating import RatingMixin
from .comment import CommentBaseModel

__all__ = [
    "TimeStampMixin",
    "UUIDBaseModel",
    "StockUnitMixin",
    "BaseTitleSlugMixin",
    "TitleSlugMixin",
    "BaseTitleSlugDescriptionMixin",
    "TitleSlugDescriptionMixin",
    "AddressMixin",
    "SingletonModelMixin",
    "RatingMixin",
    "CommentBaseModel",
]



"""
This module provides a custom CSRF middleware that extends
Django's default CsrfViewMiddleware.
It allows for the use of a custom name for the CSRF token,
enhancing the flexibility of CSRF protection in Django applications.
The primary class provided is SecureCsrfMiddleware, which can be
used as a drop-in replacement for Django's default CSRF middleware.
"""

from django.conf import settings
from django.http import FileResponse
from django.middleware.csrf import CsrfViewMiddleware as DjangoCsrfViewMiddleware
from django.utils.safestring import mark_safe


class SecureCsrfMiddleware(DjangoCsrfViewMiddleware):
    """
    SecureCsrfMiddleware extends the default CsrfViewMiddleware from
    Django to allow for a custom CSRF token name.
    It overrides the process_request and process_response methods
    to replace the standard 'csrfmiddlewaretoken'
    with a custom name defined in the class attribute 'secure_csrf_name'.
    This class provides a way to customize the CSRF token name for security
    or organizational preferences without altering the fundamental CSRF
    protection mechanisms.
    """

    # Define 'secure_csrf_name' as a class attribute
    secure_csrf_name = settings.CSRF_INPUT_NAME

    def process_request(self, request):
        """
        Processes incoming requests to replace the standard CSRF token
        name with the custom one.
        It checks for the presence of the custom CSRF token name in the
        POST data and, if found, replaces the default 'csrfmiddlewaretoken'
        with the custom token. This method ensures that the system recognizes
        the token under its new name during the request lifecycle.

        :param request: HttpRequest object representing the current request
        :return: The result of the parent class's process_request method
        """
        request.POST = request.POST.copy()
        if self.secure_csrf_name in request.POST:
            request.POST["csrfmiddlewaretoken"] = request.POST.pop(
                self.secure_csrf_name
            )[0]
        return super().process_request(request)

    def process_response(self, request, response):
        """
        Processes outgoing responses to replace occurrences of the default
        CSRF token name in the HTML content with the custom CSRF token name.
        It ensures that forms rendered in the response HTML use the custom
        token name, maintaining consistency with the request processing and
        providing a seamless experience.

        :param request: HttpRequest object representing the current request
        :param response: HttpResponse object representing the current response
        :return: The result of the parent class's process_response method with
        modified content
        """
        # Check if the response is a FileResponse instance
        if isinstance(response, FileResponse):
            # If it's a FileResponse, return it unmodified
            return response

        # If it's not a FileResponse, proceed with your original processing
        if hasattr(response, "content"):
            response.content = response.content.replace(
                b'name="csrfmiddlewaretoken"',
                mark_safe(f'name="{self.secure_csrf_name}"').encode(),
            )
        return super().process_response(request, response)



import uuid

from django.contrib import admin
from django.db import models
from django.utils.translation import gettext_lazy as _

from sage_tools.services.slug import SlugService


class TimeStampMixin(models.Model):
    """
    A mixin class that adds self-updating `created_at` and `modified_at` fields to a
    Django model.

    This mixin automatically records the creation time of an object and updates the
    modification time
    every time the object is saved. It is an abstract base class, meaning it does not
    correspond to a database table. Instead, its fields are added to each model that
    inherits from it.

    ```
    class MyModel(TimeStampMixin, models.Model):
        # other fields here
    ```

    This will add `created_at` and `modified_at` fields to `MyModel`.

    The class is marked as abstract in Django, which means it will not be used to create
    any database table. Instead, its fields are added to those of the child class.

    """

    created_at = models.DateTimeField(_("Created at"), auto_now_add=True)
    modified_at = models.DateTimeField(_("Modified at"), auto_now=True)

    class Meta:
        """
        Meta
        """

        abstract = True


class UUIDBaseModel(models.Model):
    """
    Abstract Django Model that contains:
        uuid
    """

    uuid = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        unique=True,
        help_text=_(
            "A universally unique identifier (UUID) automatically"
            " generated for this content. This identifier is used"
            " as the unique primary key and should not be modified manually."
        ),
    )

    class Meta:
        abstract = True


class StockUnitMixin(models.Model):
    """
    Abstract Django Model that contains:
        sku
    """

    sku = models.UUIDField(
        _("Stock of Unit"),
        max_length=100,
        default=uuid.uuid4,
        unique=True,
        help_text=_(
            "A unique identifier assigned to each product in inventory. "
            "This Stock Unit SKU is used to uniquely identify and manage"
            " products in your inventory."
        ),
    )

    class Meta:
        abstract = True


class BaseTitleSlugMixin(models.Model):
    """
    A mixin class that adds a title and slug field to a model.

    The `TitleSlugMixin` provides a standardized way to include title and slug fields in
    various models. It includes a `title` field, which is a CharField with a maximum
    length of 255 characters and is unique across the model it's used in. The `slug`
    field is a SlugField used for URL-friendly representations of the `title`, also
    unique and limited to 255 characters.
    """

    title = models.CharField(
        _("Title"),
        max_length=255,
        unique=True,
        help_text=_("Enter a unique title."),
        db_comment="Stores the unique title of the instance.",
    )

    slug = models.SlugField(
        _("Slug"),
        max_length=255,
        unique=True,
        editable=True,
        allow_unicode=True,
        help_text=_("URL-friendly slug from the title."),
        db_comment="Stores the URL-friendly slug derived from the title.",
    )

    @admin.display(description=_("title"), ordering=("-title"))
    def get_title(self) -> str:
        """
        Returns the title of the instance, shortened to 30 characters if necessary.

        This method is used to display a shortened version of the title in the Django
        admin interface if the title is longer than 30 characters. It appends ellipses
        to indicate that the title has been truncated.

        Returns:
            str: The full title if it's less than 30 characters, otherwise the first 30
            characters followed by '...'.
        """
        TRUNCATE_SIZE = 30
        return (
            self.title if len(self.title) < TRUNCATE_SIZE else (self.title[:30] + "...")
        )

    class Meta:
        """The Meta class of the TitleSlugMixin has an attribute abstract = True,
        making it an abstract class and the fields defined in it will be used in the
        child classes. The Meta class does not have any other attributes.
        """

        abstract = True

    def __str__(self) -> str:
        """
        String representation of the instance.

        Returns the title of the instance, providing a readable representation of the
        instance in admin panels or debug outputs.
        """
        return self.title


class TitleSlugMixin(BaseTitleSlugMixin):
    """
    A mixin class that adds a title and slug field to a model.

    The `TitleSlugMixin` provides a standardized way to include title and slug fields in
    various models. It includes a `title` field, which is a CharField with a maximum
    length of 255 characters and is unique across the model it's used in. The `slug`
    field is a SlugField used for URL-friendly representations of the `title`, also
    unique and limited to 255 characters.
    """

    def save(self, *args, **kwargs):
        slug_service = SlugService(self)
        self.slug = slug_service.create_unique_slug()
        super().save(*args, **kwargs)

    class Meta:
        """The Meta class of the TitleSlugMixin has an attribute abstract = True,
        making it an abstract class and the fields defined in it will be used in the
        child classes. The Meta class does not have any other attributes.
        """

        abstract = True

    def __str__(self) -> str:
        """
        String representation of the instance.

        Returns the title of the instance, providing a readable representation of the
        instance in admin panels or debug outputs.
        """
        return self.title


class BaseTitleSlugDescriptionMixin(BaseTitleSlugMixin):
    """
    A mixin class that extends TitleSlugMixin by adding a description field.

    This mixin inherits from TitleSlugMixin and adds a `description` field. It's
    designed to be used in models where an additional descriptive text field is
    necessary along with the title and slug. The `description` field is a TextField,
    suitable for longer text.
    """

    description = models.TextField(
        _("Description"),
        help_text=_(
            "Enter a detailed description of the item. This can include its purpose, "
            "characteristics, and any other relevant information."
        ),
        db_comment="Stores a detailed description of the instance.",
    )

    class Meta:
        """
        Meta class for TitleSlugDescriptionMixin.
        """

        abstract = True


class TitleSlugDescriptionMixin(TitleSlugMixin):
    """
    A mixin class that extends TitleSlugMixin by adding a description field.

    This mixin inherits from TitleSlugMixin and adds a `description` field. It's
    designed to be used in models where an additional descriptive text field is
    necessary along with the title and slug. The `description` field is a TextField,
    suitable for longer text.
    """

    description = models.TextField(
        _("Description"),
        help_text=_(
            "Enter a detailed description of the item. This can include its purpose, "
            "characteristics, and any other relevant information."
        ),
        db_comment="Stores a detailed description of the instance.",
    )

    class Meta:
        """
        Meta class for TitleSlugDescriptionMixin.
        """

        abstract = True



"""
Useful Functions
"""

import base64
import os
import secrets


def create_directories(base_dir: str, directories: list[tuple[str]]):
    """
    Creates a series of directories based on the input parameters.

    Args:
    - base_dir (str): The base directory in which the subdirectories will be created.
    - directories (list of tuples): A list of tuples where each tuple represents a
        subdirectory to create. The first element of each tuple is the name of the
        subdirectory, and the rest of the elements (if any) represent subdirectories
        within the previous directory.

    Example:
    base_dir = '/path/to/base'
    directories = [('logs',), ('logs', 'auth'), ('logs', 'core')]
    create_directories(base_dir, directories)

    This will create the following directories:
    - /path/to/base/logs
    - /path/to/base/logs/auth
    - /path/to/base/logs/core
    """
    for directory in directories:
        path = os.path.join(base_dir, *directory)
        if not os.path.exists(path):
            os.makedirs(path)


def generate_base32_secret(length: int = 20) -> str:
    """
    Generate a secure random Base32-encoded secret.

    Args:
        length (int): Length of the random bytes. Default is 20 bytes.

    Returns:
        str: A Base32-encoded string.
    """
    # Generate secure random bytes
    random_bytes = secrets.token_bytes(length)

    # Encode the bytes to Base32 and decode to a string for ease of use
    base32_secret = base64.b32encode(random_bytes).decode("utf-8")

    # Remove any Base32 padding characters ('=')
    return base32_secret.rstrip("=")


def hide_phone_number(phone_number):
    """
    Mask the phone number to only show the last four digits.

    Args:
        phone_number (str): The phone number to be masked.

    Returns:
        str: The masked phone number with only the last four digits visible.
    """
    if not phone_number:
        return "Phone number not available"

    # Ensure the phone number is in string format
    phone_number = str(phone_number)

    # Check if the phone number length is less than 4
    if len(phone_number) < 4:
        return "Invalid phone number"

    hidden_part = "*" * (
        len(phone_number) - 4
    )  # Replace all but the last 4 digits with asterisks
    visible_part = phone_number[-4:]  # Last 4 digits of the phone number

    return f"{hidden_part}{visible_part}"



"""
Decorators
"""

from functools import wraps

from django.conf import settings
from django.http import HttpRequest


class AdminBypassDecorator:
    """
    A decorator class designed to wrap around Django view functions or context processors.
    Its primary purpose is to prevent the execution of the wrapped function and return an
    empty context dictionary when the incoming request targets an admin page. This is
    particularly useful for avoiding unnecessary processing and ensuring sensitive
    information or functionality isn't inadvertently exposed to admin interfaces.
    """

    EMPTY_CONTEXT = {}

    def __init__(self, func):
        """
        Initializes the decorator with the function to be decorated.

        Parameters:
            func (callable): The function that this decorator will wrap. It should be
                a Django view function or context processor that takes an HttpRequest
                object as its first parameter.
        """
        wraps(func)(self)
        self.func = func

    @staticmethod
    def is_admin_request(request: HttpRequest) -> bool:
        """
        Determines if the incoming request is targeting the Django admin page.

        Parameters:
            request (HttpRequest): The incoming request object from Django.

        Returns:
            bool: True if the request is for an admin page, False otherwise. It checks
                this by comparing the request path to the admin URL name configured in
                the Django settings (settings.DJANGO_ADMINISTRATOR_URL_NAME).
        """
        return request.path.startswith(f"/{settings.DJANGO_ADMIN_URL_PREFIX}/")

    def __call__(self, request: HttpRequest, *args, **kwargs):
        """
        Executes the decorator logic and the wrapped function. If the incoming request
        is for an admin page, it bypasses the function call and returns an empty context.
        Otherwise, it calls the original function with the given arguments and keyword
        arguments.

        Returns:
            dict: The context dictionary from the wrapped function or EMPTY_CONTEXT if
                the request is for an admin page.
        """
        if self.is_admin_request(request):
            return self.EMPTY_CONTEXT
        return self.func(request, *args, **kwargs)
